//License:
//--------
//Except for the unwarp and spatial algorithm (for which the source files have
//been replaced by the object library: unwarp.lib) and the sources provided by
//the Open JPEG Foundation (see folder jpeg), all the sources contained in this
//ZIP distribution are non-proprietary information. For example, Bruker AXS Inc.
//considers all Bruker data file formats to be non-proprietary information. You
//are free to use the Bruker supplied public domain sources in any way you see
//fit.


#include <stdio.h>         /* _lseek, finddata, etc. */
#include <stdlib.h>
#include <string.h>
#ifdef _MSC_VER
#include <io.h>
typedef __int64 INT64;
#else
#include <unistd.h>
typedef unsigned long long int INT64;
#endif

#define MAXOVF 4096
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;


/* C headers */
#include <errno.h>      /* EACCES, ENOENT, ENOSPC */
#include <math.h>       /* log() */

extern "C" 
	{
	/* Bruker headers */
    #include "FRM_LIB.H"
	}

// --------------------------------------------------------------------
// ReadFrameFile: Read any Bruker frame file format (FORMAT 86 or 100).
//
// Input:   frmfile filespec
//
// Output:  frmhdr  frame header structure
//          long*   frame data (expanded to 32-bit (unsigned long))
//
// Note:    Returns NULL on error.
//
long* ReadFrameFile(char *frmfile, struct FRM_HEADER *frmhdr)
{
    long *frmdata = NULL;
    char newline[128];  // was 81
    char *frmover = NULL;   /* overflow table */
    void *frmunder = NULL;  /* underflow table */
    WORD *frmover1 = NULL;  /* over1flow table */
    DWORD *frmover2 = NULL; /* over2flow table */
    int  in_frm;        /* IN: filehandle for FRM file */
    long ier;           /* error code, 0=none */
    float linear[2];    /* linear scaling and offset of frame data */
    long nb;            /* number of bytes */
    long nbytespercount;/* number of bytes per count */
    long ncounts;       /* number of counts in image */
    long nrows, ncols;  /* number of rows and columns in image */
    long frm_version;   /* 0=Unknown, 1=Bruker, 6=Bruker/BIS */
    long novfl;         /* number of overflows */
    long novflsize = 0;
    long nunder = 0, nover1 = 0, nover2 = 0;
    long nbytesunder = 0;
    long nskip;

/* ------------------------------------------------------------------------- */
/* Open, Read, & Close input frame file                                      */
/* ------------------------------------------------------------------------- */
    /* open input FRM file, also determine FRM file format */
    long nformat;
    frm_version = 0;        // Bruker unknown
    in_frm = frm_open_file(frmfile, 0, &nformat, &ier);
    if (ier == -1) {
        printf("\aFRM2FRM-F-unknown file format (must be Bruker format)\n");
        return NULL;
    }
    if (nformat == 86) frm_version = 1;     // Bruker old
    if (nformat == 100) frm_version = 6;    // Bruker new
    if ((ier == 0) && (frm_version == 0)) {
        printf("\aFRM2FRM-F-unknown Bruker file format (not 86 or 100)\n");
        ier = frm_close_file(in_frm);
        return NULL;
    }
    if (ier == EACCES) printf("ERROR:Access or sharing violation\n");
    if (ier == ENOENT) printf("ERROR:File or path not found\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-in frm open: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-opened input frm open\n");

/* ------------------------------------------------------------------------- */
    /* read frame header from input FRM file */
    ier = frm_read_header(in_frm, frmhdr, &nb);
    if (ier != 0) {printf("\a\nFRM2FRM-F-in frm read: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-read frm header block\n");

/* ------------------------------------------------------------------------- */
    /* read image data from input FRM file */
    nbytespercount = atoi(&frmhdr->line[39][8]);
    nrows = atoi(&frmhdr->line[40][8]);
    ncols = atoi(&frmhdr->line[41][8]);
    ncounts = nrows * ncols;
    frmdata = (long *) malloc((int) ncounts * 4);    // big enough for 32-bit frames
    if (frmdata == NULL) {
        printf("Image dimensions: rows: %d cols: %d", nrows, ncols);
        printf("\a\nFRM2FRM-F-out of memory->too many counts in image\n");
        goto BAGIT;
    }
    ier = frm_read_data(in_frm, ncounts, nbytespercount, frmdata, &nb);
    if (ier != 0)
        {printf("\a\nFRM2FRM-F-in frm read data: %d\n", ier); goto BAGIT;}
    if ((nb < ncounts * nbytespercount) && (frm_version != 16))
        {printf("\nFRM2FRM-W-possible incomplete read of image data block: %d\n", nb);}
    if (dbglevel() > 2) printf("FRM2FRM-I-read image data block\n");

/* ------------------------------------------------------------------------- */
    /* read overflow table or overflow/underflow table(s) */
    switch (frm_version) {
    case 1:     /* BRUKER frame file */
        novfl = atoi(&frmhdr->line[20][8]);
        if (novfl > 0) {
            novflsize = novfl * 16;
            frmover = (char *) malloc((int) novflsize);
            if (frmover == NULL) {
                printf("\a\nFRM2FRM-F-out of memory->too many overflows\n");
                goto BAGIT;
            }
            ier = frm_read_overflow(in_frm, novflsize, frmover, &nb);
            if (ier != 0) {printf("\a\nFRM2FRM-F-in frm read overflow table: %d\n", ier); goto BAGIT;}
        }
        if (dbglevel() > 2) printf("FRM2FRM-I-read overflow table block\n");
        break;
    case 6:     /* BRUKER frame file */
        sscanf(&frmhdr->line[20][8], "%d %d %d", &nunder, &nover1, &nover2);
        if (nunder > 0) {
            sscanf(&frmhdr->line[39][8], "%d %d", &nskip, &nbytesunder);
            if (nbytesunder == 0) nbytesunder = 1;
            novflsize = nunder * nbytesunder;
            novflsize = ((novflsize+15) / 16) * 16;  // Read in 16 byte blocks
            frmunder = malloc((int) novflsize);
            if (frmunder == NULL) {
                printf("\a\nFRM2FRM-F-out of memory->too many underflows\n");
                goto BAGIT;
            }
            ier = frm_read_overflow(in_frm, novflsize, (char *) frmunder, &nb);
            if (ier != 0) {printf("\a\nFRM2FRM-F-in frm read underflow table: %d\n", ier); goto BAGIT;}
            if (dbglevel() > 2) printf("FRM2FRM-I-read underflow table block\n");
        }
        if (nover1 > 0) {
            novflsize = nover1 * 2;
            novflsize = ((novflsize+15) / 16) * 16;  // Read in 16 byte blocks
            frmover1 = (unsigned short *) malloc((int) novflsize);
            if (frmover1 == NULL) {
                printf("\a\nFRM2FRM-F-out of memory->too many over1flows\n");
                goto BAGIT;
            }
            ier = frm_read_overflow(in_frm, novflsize, (char *) frmover1, &nb);
            if (ier != 0) {printf("\a\nFRM2FRM-F-in frm read over1flow table: %d\n", ier); goto BAGIT;}
            if (dbglevel() > 2) printf("FRM2FRM-I-read over1flow table block\n");
        }
        if (nover2 > 0) {
            novflsize = nover2 * 4;
            novflsize = ((novflsize+15) / 16) * 16;  // Read in 16 byte blocks
            frmover2 = (unsigned long *) malloc((int) novflsize);
            if (frmover2 == NULL) {
                printf("\a\nFRM2FRM-F-out of memory->too many over2flows\n");
                goto BAGIT;
            }
            ier = frm_read_overflow(in_frm, novflsize, (char *) frmover2, &nb);
            if (ier != 0) {printf("\a\nFRM2FRM-F-in frm read over2flow table: %d\n", ier); goto BAGIT;}
            if (dbglevel() > 2) printf("FRM2FRM-I-read over2flow table block\n");
        }
        break;
    default:
        novfl = 0;
        break;
    }

/* ------------------------------------------------------------------------- */
    /* expand frame to 32-bit image */
    if (nbytespercount == 1) {
        unsigned char *p;
        long j, k, n;
        p = (unsigned char *) frmdata;
        k = nrows * ncols;
        for (j=0; j<nrows; j=j+1) {
            for (n=0; n<ncols; n=n+1) {
                k = k - 1;
                frmdata[k] = (long) p[k];
            }
        }
        nbytespercount = 4;
        if (frm_version == 6)
            sprintf(newline, "NPIXELB:%-35ld%-35ld ", nbytespercount, 0);
        else
            sprintf(newline, "NPIXELB:%10d", nbytespercount);
        frm_modify_header(frmhdr->line[39], newline);
    }
    else if (nbytespercount == 2) {
        unsigned short *p;
        long j, k, n;
        p = (unsigned short *) frmdata;
        k = nrows * ncols;
        for (j=0; j<nrows; j=j+1) {
            for (n=0; n<ncols; n=n+1) {
                k = k - 1;
                frmdata[k] = (long) p[k];
            }
        }
        nbytespercount = 4;
        if (frm_version == 6)
            sprintf(newline, "NPIXELB:%-35ld%-35ld ", nbytespercount, 0);
        else
            sprintf(newline, "NPIXELB:%10d", nbytespercount);
        frm_modify_header(frmhdr->line[39], newline);
    }

    /* handle overflow table */
    if ((frm_version == 1) && (frmover != NULL)) {
        long n;
        for (n=0; n<novfl; n++) {
            char strcounts[10];
            char stroffset[8];
            long offset, counts;
            strncpy(strcounts, frmover + 16 * n + 0, 9);
            strcounts[9] = '\0';
            strncpy(stroffset, frmover + 16 * n + 9, 7);
            stroffset[7] = '\0';
            counts = atoi(strcounts);
            offset = atoi(stroffset);
            frmdata[offset] = counts;
        }
        free(frmover);
    }
    else if (frm_version == 6) {
        long n;
        long n0 = 0;
        long n1 = 0;
        long n2 = 0;
        long nbaseline = 0, nver;
        sscanf(&frmhdr->line[1][8], "%d", &nver);
        if (nver > 8)
            sscanf(&frmhdr->line[79][8], "%d %d %d", &nskip, &nskip, &nbaseline);
        for (n=0; n<nrows*ncols; n++) {

            // Handle all overflows
            if ((frmover1 != NULL) && (frmdata[n] == 255)) {
                frmdata[n] = frmover1[n1];
                n1++;
            }
            if ((frmover2 != NULL) && (frmdata[n] == 65535)) {
                frmdata[n] = frmover2[n2];
                n2++;
            }

            // Handle underflow or baseline or neither
            if ((nunder > 0) && (frmdata[n] == 0)) {
                if (nbytesunder == 1) {
                    BYTE *ptr = (BYTE *) frmunder;
                    frmdata[n] = ptr[n0];
                }
                else if (nbytesunder == 2) {
                    WORD *ptr = (WORD *) frmunder;
                    frmdata[n] = ptr[n0];
                }
                else {
                    DWORD *ptr = (DWORD *) frmunder;
                    frmdata[n] = ptr[n0];
                }
                n0++;
            }
            else if (nunder != -1)
                frmdata[n] = frmdata[n] + nbaseline;
        }
        if (frmunder != NULL) free(frmunder);
        if (frmover1 != NULL) free(frmover1);
        if (frmover2 != NULL) free(frmover2);
    }

    /* handle linear scaling feature */
    sscanf(&frmhdr->line[58][8], "%f %f", &linear[0], &linear[1]);
    //linear[0] = atoi(&frmhdr->line[58][8]);
    //linear[1] = atoi(&frmhdr->line[58][8+14]);
    if (linear[0] != 1 || linear[1] != 0) {
        long k, j, n;
        k = 0;
        for (j=0; j<nrows; j=j+1) {
            for (n=0; n<ncols; n=n+1) {
                frmdata[k] = (long) (frmdata[k] * linear[0] + linear[1] + 0.5);
                k = k + 1;
            }
        }
        linear[0] = 1.0f;
        linear[1] = 0.0f;
        if (frm_version == 6)
            sprintf(newline, "LINEAR :%-35lf %-35lf", linear[0], linear[1]);
        else
            sprintf(newline, "LINEAR :%14.7f%14.7f", linear[0], linear[1]);
        frm_modify_header(frmhdr->line[58], newline);
    }

/* ------------------------------------------------------------------------- */
    /* close input FRM file */
    ier = frm_close_file(in_frm);
    if (ier != 0) goto BAGIT;

/* normal return */
    return frmdata;

/* error return */
BAGIT:
    close(in_frm);
    if (frmdata)
        free(frmdata);
    return NULL;
}

extern long SaveFrameFile_Old(char *outfile, struct FRM_HEADER *frmhdr, long *frmdata);
extern long SaveFrameFile_New(char *outfile, struct FRM_HEADER *frmhdr, long *frmdata);

// --------------------------------------------------------------------
// SaveFrameFile: Save any Bruker frame file format (FORMAT 86 or 100).
//
// Input:   frmfile filespec
//          frmhdr  frame header structure
//          frmdata frame data (expanded to 32-bit (unsigned long))
//
// Output:  Returns non-zero on error.
//          frmdata is not preserved!
//
// Note: This version always writes FORMAT 86 frame files.
//
long SaveFrameFile(char *outfile, struct FRM_HEADER *frmhdr, long *frmdata)
{
    // Always update total counts in frame header
    long nrows = atoi(&frmhdr->line[40][8]);
    long ncols = atoi(&frmhdr->line[41][8]);
    INT64 ncts = 0;
    for (int k = 0; k < nrows * ncols; k++)
        ncts += frmdata[k];

    char newline[128];
    long nFormat = atoi(&frmhdr->line[0][8]);
    INT64 nCounts[5];
    if (nFormat >= 100) {
        long num = sscanf(frmhdr->line[19]+8, "%35I64d %35I64d ", &nCounts[0],&nCounts[1]);
        nCounts[0] = ncts;
        sprintf(newline, "NCOUNTS:%-35I64d %-35I64d", nCounts[0], nCounts[1]);
    }
    else {
        long num = sscanf(frmhdr->line[19]+8, "%12d%12d%12d%12d%12d", &nCounts[0],&nCounts[1],&nCounts[2],&nCounts[3],&nCounts[4]);
        nCounts[0] = ncts;
        sprintf(newline, "NCOUNTS:%12d%12d%12d%12d%12d", nCounts[0], nCounts[1], nCounts[2], nCounts[3], nCounts[4]);
        if (num == 2)
            newline[32] = '\0';
        else if (num == 3)
            newline[44] = '\0';
        else if (num == 4)
            newline[56] = '\0';
        else if (num == 5)
            newline[68] = '\0';
    }
    frm_modify_header(frmhdr->line[19], newline);

    // Save in requested frame file format
    if (nFormat == 100)
        return SaveFrameFile_New(outfile, frmhdr, frmdata);
    else if (nFormat == 86)
        return SaveFrameFile_Old(outfile, frmhdr, frmdata);
    else
        return 1;
}

// Save in FORMAT: 86, VERSION: %d, fixed format headers
long SaveFrameFile_Old(char *outfile, struct FRM_HEADER *frmhdr, long *frmdata)
{
    char newline[128];  // was 81
    char *frmover = NULL;   /* overflow table */
    void *frmunder = NULL;  /* underflow table */
    WORD *frmover1 = NULL;  /* over1flow table */
    DWORD *frmover2 = NULL; /* over2flow table */
    int  out_frm;       /* OUT: filehandle for FRM file */
    long ier;           /* error code, 0=none */
    long nb;            /* number of bytes */
    long nbytespercount;/* number of bytes per count */
    long ncounts;       /* number of counts in image */
    long nrows, ncols;  /* number of rows and columns in image */

    long novfl, novfl8, novfl16;   /* number of overflows */
    long j, k, n;
    long maxovf;
    long novflsize = 0;
    //long nunder = 0, nover1 = 0, nover2 = 0;

/* ------------------------------------------------------------------------- */
/* Modify frame header and data if necessary                                 */
/* ------------------------------------------------------------------------- */
    // Currently only supports older compression
    sprintf(newline, "FORMAT :%10d", 86);
    frm_modify_header(frmhdr->line[0], newline);

/* ------------------------------------------------------------------------- */
    nrows = atoi(&frmhdr->line[40][8]);
    ncols = atoi(&frmhdr->line[41][8]);
    ncounts = nrows * ncols;

/* ------------------------------------------------------------------------- */
    /* update certain items in frame header */
    novfl8 = 0;
    novfl16 = 0;
    k = 0;
    for (j=0; j<nrows; j=j+1) {
        for (n=0; n<ncols; n=n+1) {
            if (frmdata[k] > 255) novfl8++;
            if (frmdata[k] > 65535) novfl16++;
            k = k + 1;
        }
    }
    //sprintf(newline, "NOVER64:%10d", novfl16);
    //frm_modify_header(frmhdr->line[38], newline);

/* ------------------------------------------------------------------------- */
    /* compress frame image from 32-bit to 8- or 16-bit */
    novfl = 0;
    maxovf = MAXOVF;
    if (getenv("SAXI$512MAXOVF")) maxovf = 512;
    if (novfl8 < maxovf) {
        unsigned char *p;
        long j, k, m, n;
        novflsize = 16l * novfl8;
        novflsize = ((novflsize + 511l) / 512l) * 512l;
        if (novflsize < 512) novflsize = 512;   // minimum overflow table size
        frmover = (char *) malloc((int) novflsize + 1); // plus room for trailing null
        if (frmover == NULL) {
            printf("\a\nFRM2FRM-F-out of memory->too many overflows\n");
            goto BAGIT;
        }
        memset(frmover, 0, novflsize);
        p = (unsigned char *) frmdata;
        k = 0;
        m = 0;
        for (j=0; j<nrows; j=j+1) {
            for (n=0; n<ncols; n=n+1) {
                if (frmdata[k] > 255) {
                    sprintf(frmover+m*16, "%9d%7d", frmdata[k], k);
                    m = m + 1;
                    frmdata[k] = 255;
                }
                p[k] = (unsigned char) frmdata[k];
                k = k + 1;
            }
        }
        novfl = novfl8;
        sprintf(newline, "NOVERFL:%10d", novfl);
        frm_modify_header(frmhdr->line[20], newline);

        nbytespercount = 1;
        sprintf(newline, "NPIXELB:%10d", nbytespercount);
        frm_modify_header(frmhdr->line[39], newline);
    }
    else if (novfl16 < maxovf) {
        unsigned short *p;
        long j, k, m, n;
        novflsize = 16l * novfl16;
        novflsize = ((novflsize + 511l) / 512l) * 512l;
        if (novflsize < 512) novflsize = 512;   // minimum overflow table size
        frmover = (char *) malloc((int) novflsize + 1); // plus room for trailing null
        if (frmover == NULL) {
            printf("\a\nFRM2FRM-F-out of memory->too many overflows\n");
            goto BAGIT;
        }
        memset(frmover, 0, novflsize);
        p = (unsigned short *) frmdata;
        k = 0;
        m = 0;
        for (j=0; j<nrows; j=j+1) {
            for (n=0; n<ncols; n=n+1) {
                if (frmdata[k] > 65535l) {
                    sprintf(frmover+m*16, "%9d%7d", frmdata[k], k);
                    m = m + 1;
                    frmdata[k] = 65535l;
                }
                p[k] = (unsigned short) frmdata[k];
                k = k + 1;
            }
        }
        novfl = novfl16;
        sprintf(newline, "NOVERFL:%10d", novfl);
        frm_modify_header(frmhdr->line[20], newline);

        nbytespercount = 2;
        sprintf(newline, "NPIXELB:%10d", nbytespercount);
        frm_modify_header(frmhdr->line[39], newline);
    }
    else {
        // Store as 4-byte per pixel image
        novfl = 0;
        sprintf(newline, "NOVERFL:%10d", novfl);
        frm_modify_header(frmhdr->line[20], newline);

        nbytespercount = 4;
        sprintf(newline, "NPIXELB:%10d", nbytespercount);
        frm_modify_header(frmhdr->line[39], newline);

        // Note: LINEAR compression could be used here to make 2-byte frame
        // On CCD frames only, try to create a 2-byte frame without overflows
        // by right-shifting all pixels by X bits. Restricted to upto 3 bits.
        // (Preserve some negative counts for statistical purposes).
        long j, k = 0, n;
        long nmaxpixelval = 0;
        for (j=0; j<nrows; j=j+1) {
            for (n=0; n<ncols; n=n+1) {
                if (frmdata[k] > nmaxpixelval)
                    nmaxpixelval = frmdata[k];
                k = k + 1;
            }
        }
        long nbaseline = 0, nskip, nver;
        sscanf(&frmhdr->line[1][8], "%d", &nver);
        if (nver > 8)
            sscanf(&frmhdr->line[79][8], "%d %d %d", &nskip, &nskip, &nbaseline);
        if ((nbaseline > 0) && (nmaxpixelval > 65535) && (nmaxpixelval < 524288)) {
            float linear[2] = {1.0, 0.0};
            int irightshift = int( log( float(nmaxpixelval)/65536.0) / log(2.0) + 0.9999);
            long j, k = 0, n;
            unsigned short *p;
            p = (unsigned short *) frmdata;
            for (j=0; j<nrows; j=j+1) {
                for (n=0; n<ncols; n=n+1) {
                    p[k] = (unsigned short) (frmdata[k] >> irightshift);
                    k = k + 1;
                }
            }

            linear[0] = (float) pow(2.0f, (float) irightshift);
            sprintf(newline, "LINEAR :%14.7f%14.7f", linear[0], linear[1]);
            frm_modify_header(frmhdr->line[58], newline);

            nbytespercount = 2;
            sprintf(newline, "NPIXELB:%10d", nbytespercount);
            frm_modify_header(frmhdr->line[39], newline);
        }
    }

/* ------------------------------------------------------------------------- */
/* Open, Write, & Close output frame file                                    */
/* ------------------------------------------------------------------------- */
    /* pre-delete output file */
    ier = remove(outfile);

    /* open output FRM file */
    out_frm = frm_open_file(outfile, 1, NULL, &ier);

    if (ier == EACCES) printf("ERROR:Access or sharing violation\n");
    if (ier == ENOENT) printf("ERROR:File or path not found\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm open: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-opened output frm open\n");

/* ------------------------------------------------------------------------- */
    /* write header to output FRM file */
    ier = frm_write_header(out_frm, frmhdr, &nb);
    if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-wrote frm header block\n");

    if (dbglevel() > 0) {
        /* output selected file header information */
        //printf("%80s\n", frmhdr.line[11][0]);
    }

/* ------------------------------------------------------------------------- */
    /* write image data to output FRM file */
    ier = frm_write_data(out_frm, ncounts, nbytespercount, frmdata, &nb);
    if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write data: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-wrote image data block\n");

/* ------------------------------------------------------------------------- */
    /* write overflow table */
    ier = frm_write_overflow(out_frm, novflsize, frmover, &nb);
    if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write overflow table: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-wrote overflow table block\n");
    if (frmover != NULL) free(frmover);

/* ------------------------------------------------------------------------- */
    /* close output FRM file */
    ier = frm_close_file(out_frm);
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm close: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-closed output frm file\n");

    /* exit with error status */
    //free(frmdata);
    //frmdata = NULL;
    return 0;

/* error return */
BAGIT:
    close(out_frm);
    /* delete the failed or partially saved file, if any */
    if (dbglevel() == 0) ier = remove(outfile);
    return 1;
}

// Save in FORMAT: 100, VERSION: %d, free format header
long SaveFrameFile_New(char *outfile, struct FRM_HEADER *frmhdr, long *frmdata)
{
    char newline[128];  // was 81
    void *frmunder = NULL;  /* underflow table */
    WORD *frmover2 = NULL;  /* over1flow table */
    DWORD *frmover4 = NULL; /* over2flow table */
    int  out_frm;       /* OUT: filehandle for FRM file */
    long ier;           /* error code, 0=none */
    long nb;            /* number of bytes */
    long ncounts;       /* number of counts in image */

    long nrows, ncols;  /* number of rows and columns in image */
    long npixelb[2];    /* nbytes per pixel (image, underflow) */
    long novfl[3];      /* number of pixels in (under, over8, over16) */

    long nunder, novfl8, novfl8wu, novfl16, novfl16wu;/* number of overflows */
    long j, k, n;
    long novflsize = 0;
    long nbaseline, nskip;
    long nbytesunder, nbytesover8, nbytesover8wu, nbytesover16, nbytesover16wu;

/* ------------------------------------------------------------------------- */
/* Modify frame header and data if necessary                                 */
/* ------------------------------------------------------------------------- */
    // Currently only supports newer compression
    sprintf(newline, "FORMAT :%-72ld", 100);
    frm_modify_header(frmhdr->line[0], newline);

/* ------------------------------------------------------------------------- */
    sscanf(&frmhdr->line[79][8], "%d %d %d", &nskip, &nskip, &nbaseline);
    if (nbaseline > 65353)
        npixelb[1] = 4;
    else if (nbaseline > 255)
        npixelb[1] = 2;
    else if (nbaseline > 0)
        npixelb[1] = 1;
    else
        npixelb[1] = 0;

    nrows = atoi(&frmhdr->line[40][8]);
    ncols = atoi(&frmhdr->line[41][8]);
    ncounts = nrows * ncols;

/* ------------------------------------------------------------------------- */
    /* update certain items in frame header */
    nunder = 0;
    novfl8 = 0;
    novfl16 = 0;
    novfl8wu = 0;
    novfl16wu = 0;
    k = 0;
    for (j=0; j<nrows; j=j+1) {
        for (n=0; n<ncols; n=n+1) {
            if (frmdata[k] <= nbaseline) nunder++;
            if (frmdata[k] >= 255) novfl8++;
            if (frmdata[k] >= 65535) novfl16++;
            if (frmdata[k] - nbaseline >= 255) novfl8wu++;
            if (frmdata[k] - nbaseline >= 65535) novfl16wu++;
            k = k + 1;
        }
    }
    if (nbaseline == 0) nunder = 0;

    /* determine smallest compression method */
    nbytesunder = nunder * npixelb[1];
    nbytesunder = ((nbytesunder + 15) / 16) * 16;

    nbytesover8 = novfl8 * 2;
    nbytesover8 = ((nbytesover8 + 15) / 16) * 16;

    nbytesover8wu = novfl8wu * 2;
    nbytesover8wu = ((nbytesover8wu + 15) / 16) * 16;

    nbytesover16 = novfl16 * 4;
    nbytesover16 = ((nbytesover16 + 15) / 16) * 16;

    nbytesover16wu = novfl16wu * 4;
    nbytesover16wu = ((nbytesover16wu + 15) / 16) * 16;

    //  [0] -- Uncompressed 4-byte pixels, (nrows * ncols * 4)
    //  [1] -- 1-byte pixels, 16- & 32-bit overflows
    //  [2] -- 2-byte pixels, 32-bit overflows
    //  [3] -- 1-byte pixels, underflows & baseline subtraction, 16- & 32- bit overflow compression
    //  [4] -- 2-byte pixels, underflows & baseline subtraction, 32- bit overflow compression
    long nSize[5], nBest;
    nSize[0] = ncounts * 4;
    nBest = 0;
    nSize[1] = ncounts + nbytesover8 + nbytesover16;
    if (nSize[1] < nSize[nBest]) nBest = 1;
    nSize[2] = ncounts * 2 + nbytesover16;
    if (nSize[2] < nSize[nBest]) nBest = 2;
    nSize[3] = ncounts + nbytesunder + nbytesover8wu + nbytesover16wu;
    if (nSize[3] < nSize[nBest]) nBest = 3;
    nSize[4] = ncounts * 2 + nbytesunder + nbytesover16wu;
    if (nSize[4] < nSize[nBest]) nBest = 4;

    switch (nBest) {
    case 0:
        npixelb[0] = 4;
        novfl[0] = -1;
        novfl[1] = 0;
        novfl[2] = 0;
        break;
    case 1:
        npixelb[0] = 1;
        novfl[0] = -1;
        novfl[1] = novfl8;
        novfl[2] = novfl16;
        break;
    case 2:
        npixelb[0] = 2;
        novfl[0] = -1;
        novfl[1] = 0;
        novfl[2] = novfl16;
        break;
    case 3:
        npixelb[0] = 1;
        novfl[0] = nunder;
        novfl[1] = novfl8wu;
        novfl[2] = novfl16wu;
        break;
    case 4:
        npixelb[0] = 2;
        novfl[0] = nunder;
        novfl[1] = 0;
        novfl[2] = novfl16wu;
        break;
    }

    sprintf(newline, "NOVERFL:%-23ld %-23ld %-23ld ", novfl[0], novfl[1], novfl[2]);
    frm_modify_header(frmhdr->line[20], newline);
    sprintf(newline, "NPIXELB:%-35ld %-35ld ", npixelb[0], npixelb[1]);
    frm_modify_header(frmhdr->line[39], newline);

/* ------------------------------------------------------------------------- */
    /* allocate under- and over- flow table(s) */
    long nundersize = 0, nover2size = 0, nover4size = 0;
    if (novfl[0] > 0) {
        nundersize = nbytesunder;
        frmunder = malloc(nbytesunder);
        if (frmunder == NULL) {
            printf("\a\nFRM2FRM-F-out of memory->too many underflows\n");
            goto BAGIT;
        }
        memset(frmunder, 0, nbytesunder);
    }
    else 
        frmunder = NULL;

    if (novfl[1] > 0) {
        nover2size = ((novfl[1] * 2 + 15) / 16) * 16;
        frmover2 = (WORD *) malloc(nover2size);
        if (frmover2 == NULL) {
            printf("\a\nFRM2FRM-F-out of memory->too many 2-byte overflows\n");
            goto BAGIT;
        }
        memset(frmover2, 0, nover2size);
    }
    else 
        frmover2 = NULL;

    if (novfl[2] > 0) {
        nover4size = ((novfl[2] * 4 + 15) / 16) * 16;
        frmover4 = (DWORD *) malloc(nover4size);
        if (frmover4 == NULL) {
            printf("\a\nFRM2FRM-F-out of memory->too many 4-byte overflows\n");
            goto BAGIT;
        }
        memset(frmover4, 0, nover4size);
    }
    else 
        frmover4 = NULL;

    /* compress frame image from 32-bit to 8- or 16-bit with under-/over- tables */
{
    BYTE *ptrimage1 = (BYTE *) frmdata;
    WORD *ptrimage2 = (WORD *) frmdata;
    BYTE *ptrunder1 = (BYTE *) frmunder;
    WORD *ptrunder2 = (WORD *) frmunder;
    DWORD *ptrunder4 = (DWORD *) frmunder;
    WORD *ptrover2 = frmover2;
    DWORD *ptrover4 = frmover4;
    k = 0;
    for (j=0; j<nrows; j=j+1) {
        for (n=0; n<ncols; n=n+1) {

            if (frmunder) {
                if (frmdata[k] <= nbaseline) {
                    switch (npixelb[1]) {
                    case 1:
                        *ptrunder1 = (BYTE) frmdata[k];
                        ptrunder1++;
                        break;
                    case 2:
                        *ptrunder2 = (WORD) frmdata[k];
                        ptrunder2++;
                        break;
                    case 4:
                        *ptrunder4 = (DWORD) frmdata[k];
                        ptrunder4++;
                        break;
                    }
                    frmdata[k] = 0;
                }
                else
                    frmdata[k] -= nbaseline;
            }

            if (frmover4) {
                if (frmdata[k] >= 65535) {
                    *ptrover4 = frmdata[k];
                    ptrover4++;
                    frmdata[k] = 65535;
                }
            }

            if (frmover2) {
                if (frmdata[k] >= 255) {
                    *ptrover2 = (WORD) frmdata[k];
                    ptrover2++;
                    frmdata[k] = 255;
                }
            }

            switch (npixelb[0]) {
            case 1:
                *ptrimage1 = (BYTE) frmdata[k];
                ptrimage1++;
                break;
            case 2:
                *ptrimage2 = (WORD) frmdata[k];
                ptrimage2++;
                break;
            case 4:
                break;
            }
            k = k + 1;
        }
    }
    // Internal checks (for Joerge/APEX2 issues)
    if (frmunder) {}
    if (frmover2) {
        if (novfl[1] != (ptrover2 - frmover2)) {
			printf("2-byte-overflow table issue\n)");
        }
    }
    if (frmover4) {
        if (novfl[2] != (ptrover4 - frmover4)) {
			printf("4-byte-overflow table issue\n)");
        }
    }
}

/* ------------------------------------------------------------------------- */
/* Open, Write, & Close output frame file                                    */
/* ------------------------------------------------------------------------- */
    /* pre-delete output file */
    ier = remove(outfile);

    /* open output FRM file */
    out_frm = frm_open_file(outfile, 1, NULL, &ier);

    if (ier == EACCES) printf("ERROR:Access or sharing violation\n");
    if (ier == ENOENT) printf("ERROR:File or path not found\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm open: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-opened output frm open\n");

/* ------------------------------------------------------------------------- */
    /* write header to output FRM file */
    ier = frm_write_header(out_frm, frmhdr, &nb);
    if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-wrote frm header block\n");

    if (dbglevel() > 0) {
        /* output selected file header information */
        //printf("%80s\n", frmhdr.line[11][0]);
    }

/* ------------------------------------------------------------------------- */
    /* write image data to output FRM file */
    ier = frm_write_data(out_frm, ncounts, npixelb[0], frmdata, &nb);
    if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write data: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-wrote image data block\n");

/* ------------------------------------------------------------------------- */
    /* write underflow table */            /* Padded to multiple of 16-bytes */
    if (frmunder) {
        ier = frm_write_data(out_frm, nundersize/npixelb[1], npixelb[1], (long *) frmunder, &nb);
        if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
        if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write underflow table: %d\n", ier); goto BAGIT;}
        if (dbglevel() > 2) printf("FRM2FRM-I-wrote underflow table block\n");
        if (frmunder != NULL) free(frmunder);
    }

/* ------------------------------------------------------------------------- */
    /* write 2-byte overflow table */      /* Padded to multiple of 16-bytes */
    if (frmover2) {
        ier = frm_write_data(out_frm, nover2size/2, 2, (long *) frmover2, &nb);
        if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
        if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write 2-byte overflow table: %d\n", ier); goto BAGIT;}
        if (dbglevel() > 2) printf("FRM2FRM-I-wrote 2-byte overflow table block\n");
        if (frmover2 != NULL) free(frmover2);
    }

/* ------------------------------------------------------------------------- */
    /* write 4-byte overflow table */      /* Padded to multiple of 16-bytes */
    if (frmover4) {
        ier = frm_write_data(out_frm, nover4size/4, 4, (long *) frmover4, &nb);
        if (ier == ENOSPC) printf("ERROR:Not enough space on device\n");
        if (ier != 0) {printf("\a\nFRM2FRM-F-out frm write 4-byte overflow table: %d\n", ier); goto BAGIT;}
        if (dbglevel() > 2) printf("FRM2FRM-I-wrote 4-byte overflow table block\n");
        if (frmover4 != NULL) free(frmover4);
    }

/* ------------------------------------------------------------------------- */
    /* close output FRM file */
    ier = frm_close_file(out_frm);
    if (ier != 0) {printf("\a\nFRM2FRM-F-out frm close: %d\n", ier); goto BAGIT;}
    if (dbglevel() > 2) printf("FRM2FRM-I-closed output frm file\n");

    /* exit with error status */
    //free(frmdata);
    //frmdata = NULL;
    return 0;

/* error return */
BAGIT:
    close(out_frm);
    /* delete the failed or partially saved file, if any */
    if (dbglevel() == 0) ier = remove(outfile);
    return 1;
}
